// Fichier Registry.cpp

#include <string>
#include <objbase.h>
#include <assert.h>
#include <registry.h>
#include <ConversionChaines.h>

// -------------------------------------------
// -------------------------------------------
//
// Classe utilitaire pour enrober certaines oprations
//  la base de registres du systme
//
// -------------------------------------------
// -------------------------------------------


class CleRegistre
{
   HKEY m_Ruche;
   std::string m_Cle, m_SousCle, m_Valeur;
   void EcrireValeur (const HKEY hKey, const std::string &ValeurAEcrire = "")
   {
      if (m_Valeur.length ())
      {
         const char *Texte = ValeurAEcrire.length ()? ValeurAEcrire.c_str () : 0;
         ::RegSetValueEx
            (hKey, Texte, 0, REG_SZ,
            reinterpret_cast<BYTE *>(const_cast<char *>(m_Valeur.c_str ())),
            static_cast<DWORD> (m_Valeur.length () + 1));
      }
   }
public:
   static const std::string SEPARATEUR;
   class ErreurCreation { };
   class ErreurOuverture { };
   class ErreurSuppression { };
   CleRegistre (const HKEY Ruche, const std::string &Cle,
                                  const std::string &SousCle,
                                  const std::string &Valeur)
      : m_Ruche (Ruche),
        m_Cle (Cle), m_SousCle (SousCle), m_Valeur (Valeur)
   {
   }
   void Creer () throw (ErreurCreation)
   {
      std::string KeyBuf (m_Cle);
      // Ajoute nom de sous-cl au tampon.
      if(m_SousCle.length ())
         KeyBuf += SEPARATEUR + m_SousCle;
      // Cre et ouvre cl plus sous-cl.
      HKEY hKey;
      long lResult =
         ::RegCreateKeyEx
            (m_Ruche, KeyBuf.c_str (), 0, 0,
             REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 0, &hKey, 0);
      if(lResult != ERROR_SUCCESS) throw ErreurCreation();
      // dfinit la valeur.
      EcrireValeur (hKey);
      ::RegCloseKey(hKey);
   }
   void SetValeur (const std::string &ValeurNommee) throw (ErreurOuverture)
   {
      // Cre et ouvre cl plus sous-cl.
      HKEY hKey;
      long lResult =
         ::RegOpenKeyEx(m_Ruche, m_Cle.c_str (), 0, KEY_SET_VALUE, &hKey);
      if(lResult != ERROR_SUCCESS) throw ErreurOuverture();
      // dfinit la valeur
      EcrireValeur (hKey, ValeurNommee);
      ::RegCloseKey(hKey);
   }
   virtual ~CleRegistre ()
   {
   }
   //
   // supprime une cl et tous ses descendants
   //
   static long SuppressionRecursive 
      (HKEY hKeyParent, const std::string & CleEnfant) throw (ErreurSuppression)
   {
      // Ouvre l'enfant.
      HKEY hKeyEnfant;
      LONG lRes = ::RegOpenKeyEx(hKeyParent, CleEnfant.c_str (), 0, KEY_ALL_ACCESS, &hKeyEnfant);
      if(lRes != ERROR_SUCCESS) throw ErreurSuppression();

      // Enumre tous les decendents de cet enfant.
      FILETIME time;
      const int TAILLE_TAMPON = 256;
      char szBuffer[TAILLE_TAMPON];
      DWORD dwSize = TAILLE_TAMPON;
      while(::RegEnumKeyEx(hKeyEnfant, 0, szBuffer, &dwSize, 0, 0, 0, &time) == S_OK)
      {
         // supprime les decendants de cet enfant.
         lRes = SuppressionRecursive(hKeyEnfant, szBuffer);
         if(lRes != ERROR_SUCCESS)
         {
            // nettoyage avant fin.
            ::RegCloseKey(hKeyEnfant);
            throw ErreurSuppression();
         }
         dwSize = TAILLE_TAMPON;
      }
      // Ferme l'enfant.
      ::RegCloseKey(hKeyEnfant);
      // Supprime cet enfant.
      long Resultat = ::RegDeleteKey(hKeyParent, CleEnfant.c_str ()); 
      if (Resultat != ERROR_SUCCESS && Resultat != ERROR_FILE_NOT_FOUND)
         throw ErreurSuppression();
   }
};

const std::string CleRegistre::SEPARATEUR = "\\";


// -------------------------------------------
// -------------------------------------------



struct REG_DATA
{
    const char* pszKey;
    const char* pszValue;
    const char* pszData;
};

//
// Dclarations locales
//
namespace OutilsEnregistrement
{
   //
   // Convertit un CLSID en une chaine de caractres.
   //
   std::string CLSIDtochar(REFCLSID clsid);
   //
   // construit la cl CLSID\\{...}
   //
   std::string ContruireCLSID
      (const std::string & clsid_brut);
   //
   //
   //
   bool EnregistrementCle
      (const std::string & Cle, const std::string & NomConvivial);
   //
   //
   //
   bool EnregistrementProgID
      (const std::string & Cle, const std::string & ProgID);
   //
   //
   //
   bool EnregistrementTypeServeur
      (const std::string & NomModule,
       const std::string & NomModuleComplet,
       const std::string & Cle,
       const char * ThreadingModel);
   //
   //
   //
   bool EnregistrementVersionDependent
      (const std::string & ProgID,
       const std::string & NomConvivial,
       const std::string & clsid_brut);
   //
   //
   //
   bool EnregistrementCleVersionIndependent
      (const std::string & Cle,
       const std::string & NomIndependantDeVersion);
   //
   //
   //
   bool EnregistrementVersionIndependent
      (const std::string & ProgID,
       const std::string & NomConvivial,
       const std::string & clsid_brut,
       const std::string & NomIndependantDeVersion);
   //
   // Dfinit la cl donne et sa valeur.
   //
   bool FixerCleEtValeur
      (const std::string & Cle,
       const std::string & SousCle,
       const std::string & Valeur);
   //
   // Ouvre une cl et dfinit une valeur.
   //
   bool FixerValeurDansCle
      (const std::string & Cle,
       const std::string & ValeurNommee,
       const std::string & Valeur);
   //
   // Obtenir le nom complet d'un module (chemin inclus)
   //
   std::string ObtenirNomCompletModule
      (HMODULE hModule);
   //
   // Inscription au registre et suppression du registre en bloc
   //
   HRESULT RegisterServerEx
      (const REG_DATA regData[], const char* szModuleName);
   HRESULT UnregisterServerEx
      (const REG_DATA regData[]);
}

//
// construit la cl CLSID\\{...}
//
std::string OutilsEnregistrement::ContruireCLSID (const std::string & clsid_brut)
{
   const std::string CLSID_PREFIXE = "CLSID\\";
   return CLSID_PREFIXE + clsid_brut;
}

//
//
//
bool OutilsEnregistrement::EnregistrementCle
   (const std::string & Cle, const std::string & NomConvivial)
{
   const std::string SOUSCLE_VIDE;
   return FixerCleEtValeur (Cle, SOUSCLE_VIDE, NomConvivial);
}

//
//
//
bool OutilsEnregistrement::EnregistrementProgID
   (const std::string & Cle, const std::string & ProgID)
{
   const std::string SOUSCLE_PROGID = "ProgID";
   return FixerCleEtValeur (Cle, SOUSCLE_PROGID, ProgID);
}

//
//
//
bool OutilsEnregistrement::EnregistrementTypeServeur (const std::string & NomModule,
                                const std::string & NomModuleComplet,
                                const std::string & Cle,
                                const char * ThreadingModel)
{
   const std::string
      EXTENSION_SERVEUR_EXTERNE = ".exe",
      PREFIXE_MODELE_MULTIPROG = "ThreadingModel";
   // Ajoute la sous-cl "nom de fichier du serveur" sous la cl CLSID.
   bool ResultatEnregistrement;
   if(NomModule.find (EXTENSION_SERVEUR_EXTERNE) == std::string::npos)
   {
      const std::string
         NOMCLE_SERVEUR_INTERNE= "InprocServer32";
      std::string CleInterne = Cle + CleRegistre::SEPARATEUR + NOMCLE_SERVEUR_INTERNE;
      ResultatEnregistrement =
         FixerCleEtValeur (Cle, NOMCLE_SERVEUR_INTERNE, NomModuleComplet) &&
         FixerValeurDansCle (CleInterne, PREFIXE_MODELE_MULTIPROG,
                             ThreadingModel? ThreadingModel : "");
   }
   else
   {
      const std::string NOMCLE_SERVEUR_EXTERNE = "LocalServer32";
      ResultatEnregistrement =
         FixerCleEtValeur (Cle, NOMCLE_SERVEUR_EXTERNE, NomModuleComplet);
   }
   return ResultatEnregistrement;
}

//
//
//
bool OutilsEnregistrement::EnregistrementVersionDependent (const std::string & ProgID,
                                     const std::string & NomConvivial,
                                     const std::string & clsid_brut)
{
   const std::string
      SOUSCLE_CLSID = "CLSID",
      SOUSCLE_VIDE;
   return FixerCleEtValeur (ProgID, SOUSCLE_VIDE, NomConvivial) &&
          FixerCleEtValeur (ProgID, SOUSCLE_CLSID, clsid_brut);
}

//
//
//
bool OutilsEnregistrement::EnregistrementCleVersionIndependent (const std::string & Cle,
                                          const std::string & NomIndependantDeVersion)
{
   const std::string
      SOUSCLE_VERSIONINDEPENDENTPROGID = "VersionIndependentProgID";
   return FixerCleEtValeur (Cle, SOUSCLE_VERSIONINDEPENDENTPROGID,
                            NomIndependantDeVersion);
}

//
//
//
bool OutilsEnregistrement::EnregistrementVersionIndependent
   (const std::string & ProgID, const std::string & NomConvivial,
    const std::string & clsid_brut, const std::string & NomIndependantDeVersion)
{
   const std::string
      SOUSCLE_CLSID = "CLSID",
      SOUSCLE_VERSION = "CurVer",
      SOUSCLE_VIDE;
   return FixerCleEtValeur (NomIndependantDeVersion,
                            SOUSCLE_VIDE, NomConvivial) && 
          FixerCleEtValeur (NomIndependantDeVersion,
                            SOUSCLE_CLSID, clsid_brut) &&
          FixerCleEtValeur (NomIndependantDeVersion,
                            SOUSCLE_VERSION, ProgID);
}

//
// Dfinit la cl donne et sa valeur.
//
bool OutilsEnregistrement::FixerCleEtValeur(const std::string & Cle, const std::string & SousCle, const std::string & Valeur)
{
   CleRegistre CleACreer(HKEY_CLASSES_ROOT, Cle, SousCle, Valeur);
   try
   {
      CleACreer.Creer ();
   }
   catch (CleRegistre::ErreurCreation)
   {
      return false;
   }
   return true;
}

//
// Ouvre une cl et dfinit une valeur.
//
bool OutilsEnregistrement::FixerValeurDansCle (const std::string & Cle, const std::string & ValeurNommee, const std::string & Valeur)
{
   CleRegistre CleAModifier(HKEY_CLASSES_ROOT, Cle, "", Valeur);
   try
   {
      CleAModifier.SetValeur (ValeurNommee);
   }
   catch (CleRegistre::ErreurOuverture)
   {
      return false;
   }
   return true;
}


//
// Obtenir le nom complet d'un module (chemin inclus)
//
std::string OutilsEnregistrement::ObtenirNomCompletModule (HMODULE hModule)
{
   // Taille d'un CLSID en tant que chaine.
   const int TAILLE_MAX_NOMFICHIER = 512;
   char szModule[TAILLE_MAX_NOMFICHIER];
   DWORD dwResult =
      GetModuleFileName(hModule, szModule, TAILLE_MAX_NOMFICHIER);
   assert(dwResult);
   return szModule;
}

// -------------------------------------------------------
// -------------------------------------------------------
// -------------------------------------------------------

//
// Inscription au registre et suppression du registre (un  la fois)
//
HRESULT RegisterServer
   (const std::string & NomModule, REFCLSID clsid,
    const std::string & NomConvivial, const std::string & NomIndependantDeVersion,
    const std::string & ProgID, const char * ThreadingModel)
{
   // obtient l'emplacement du serveur
   const std::string NomModuleComplet = OutilsEnregistrement::ObtenirNomCompletModule (GetModuleHandle(NomModule.c_str ()));
   const std::string clsid_brut = OutilsEnregistrement::CLSIDtochar (clsid);
   const std::string Cle = OutilsEnregistrement::ContruireCLSID (clsid_brut);

   // Ajoute le CLSID au registre.
   HRESULT hr = S_OK;
   if (!OutilsEnregistrement::EnregistrementCle (Cle, NomConvivial) ||
       !OutilsEnregistrement::EnregistrementTypeServeur (NomModule, NomModuleComplet, Cle, ThreadingModel) ||
       !OutilsEnregistrement::EnregistrementProgID (Cle, ProgID) ||
       !OutilsEnregistrement::EnregistrementCleVersionIndependent (Cle, NomIndependantDeVersion) ||
       !OutilsEnregistrement::EnregistrementVersionIndependent    (ProgID, NomConvivial, clsid_brut, NomIndependantDeVersion) ||
       !OutilsEnregistrement::EnregistrementVersionDependent      (ProgID, NomConvivial, clsid_brut))
      hr = S_FALSE;

   return hr;
}

// Supprime le composant du registre.
LONG UnregisterServer
   (REFCLSID clsid, const std::string & NomIndependantDeVersion,
                    const std::string & ProgID)
{
   const std::string clsid_brut = OutilsEnregistrement::CLSIDtochar (clsid);
   const std::string Cle = OutilsEnregistrement::ContruireCLSID (clsid_brut);

   try
   {
      // supprime la cl CLSID\{...}
      CleRegistre::SuppressionRecursive(HKEY_CLASSES_ROOT, Cle);
      // supprime la cl version-independent ProgID.
      CleRegistre::SuppressionRecursive(HKEY_CLASSES_ROOT, NomIndependantDeVersion);
      // supprime la cl ProgID.
      CleRegistre::SuppressionRecursive(HKEY_CLASSES_ROOT, ProgID);
   }
   catch (CleRegistre::ErreurSuppression)
   {
      return E_FAIL;
   }
   return S_OK;
}

// Convertit un CLSID en une chaine.
std::string OutilsEnregistrement::CLSIDtochar(REFCLSID clsid)
{
   // Get CLSID
   LPOLESTR wszCLSID = 0;
   HRESULT hr = StringFromCLSID(clsid, &wszCLSID);
   assert(SUCCEEDED(hr));
   const std::string clsid_obtenu = WStringToString (wszCLSID);
   // Libre la mmoire.
   ::CoTaskMemFree(wszCLSID);
   return clsid_obtenu;
}


//
// Inscription au registre et suppression du registre (en bloc)
//
HRESULT OutilsEnregistrement::RegisterServerEx
   (const REG_DATA regData[], const char* szModuleName)
{
   for(int count = 0;; count++)
   {
      const char* pszKey = regData[count].pszKey;
      const char* pszValue = regData[count].pszValue;
      const char* pszData = regData[count].pszData;

      if(pszKey == 0 && pszValue == 0 && pszData == 0)
         break;

      if(pszData == (const char*)-1)
         pszData = szModuleName;

      HKEY hKey;
      long err = RegCreateKey(HKEY_CLASSES_ROOT, pszKey, &hKey);
      if(err == ERROR_SUCCESS)
      {
         err = RegSetValueEx(hKey, pszValue, 0, REG_SZ, (const BYTE*)pszData,
                             static_cast<DWORD>(strlen(pszData) + 1));
         RegCloseKey(hKey);
      }
      if(err != ERROR_SUCCESS)
      {
         UnregisterServerEx(regData);
         return REGDB_E_WRITEREGDB;
      }
   }
   return S_OK;
}

HRESULT OutilsEnregistrement::UnregisterServerEx(const REG_DATA regData[])
{
   HRESULT hr = S_OK;

   for(int nEntries = 0;; nEntries++)
      if(regData[nEntries].pszKey == 0 && regData[nEntries].pszValue == 0 && regData[nEntries].pszData == 0)
         break;

   for(int count = nEntries - 1; count >= 0; count--)
      if(RegDeleteKeyA(HKEY_CLASSES_ROOT, regData[count].pszKey) != ERROR_SUCCESS)
         hr = S_FALSE;

   return hr;
}